# coding=utf-8
import base64
import hashlib
import os
import random
import webbrowser

import click
import logging
import requests
import sys

from flask import request, redirect
from werkzeug.serving import make_server

from .config import missing_link_config, Config

base_url = '_ah/api/missinglink/v1/'


def create_validate_length(min_length, max_length):
    def validate_length(ctx, option, value):
        if value is None:
            return value

        if len(value) > max_length:
            raise click.BadParameter("needs to be shorter than %s characters" % max_length)

        if len(value) < min_length:
            raise click.BadParameter("needs to be longer than %s characters" % min_length)

        return value

    return validate_length


def auth0_url(auth0):
    return '{}.auth0.com'.format(auth0)


def add_to_data_if_not_none(data, val, key):
    if val is not None:
        data[key] = val


def get_data_and_remove(kwargs, key):
    key = key.lower()
    result = kwargs.get(key)

    if key in kwargs:
        del kwargs[key]

    return result


def base64url(b):
    return bytearray(base64.b64encode(b).decode('ascii').replace('=', '').replace('+', '-').replace('/', '_'), 'ascii')


def sha256(s):
    h = hashlib.sha256()
    h.update(s)
    return h.digest()


def url_encode(d):
    try:
        # noinspection PyUnresolvedReferences
        from urllib.parse import urlencode
        return urlencode(d)
    except ImportError:
        import urllib
        # noinspection PyUnresolvedReferences
        return urllib.urlencode(d)


def urljoin(*args):
    try:
        # noinspection PyUnresolvedReferences
        import urlparse
        method = urlparse.urljoin
    except ImportError:
        import urllib.parse
        method = urllib.parse.urljoin

    base = args[0]
    for u in args[1:]:
        base = method(base, u)

    return base


# noinspection PyUnusedLocal
def handle_api(ctx, http_method, method_url, data=None):
    url = urljoin(ctx.api_host, base_url, method_url)

    result = None
    for retries in range(3):
        headers = {'Authorization': 'Bearer {}'.format(ctx.id_token)}
        r = http_method(url, headers=headers, json=data)

        if r.status_code == 401:
            update_token(ctx.client_id, ctx.refresh_token, ctx.auth0)
            continue

        r.raise_for_status()
        result = r.json()
        break

    if result is None:
        click.echo('failed to refresh the token, rerun auth init again', err=True)

    return result


def update_token(client_id, refresh_token, auth0):
    r = requests.post('https://{}/delegation'.format(auth0_url(auth0)), json={
        'client_id': client_id,
        'grant_type': "urn:ietf:params:oauth:grant-type:jwt-bearer",
        'scope': 'openid offline_access user_external_id',
        'refresh_token': refresh_token,
    })

    r.raise_for_status()

    data = r.json()

    config = Config()
    config.set('token', 'id_token', data['id_token'])

    with open(missing_link_config, 'w') as configfile:
        config.write(configfile)


class LocalWebServer:
    ports = [27017, 41589, 55130, 6617, 1566, 17404, 44288, 6948, 15950, 48216, 3318, 34449, 22082, 22845, 37862, 59304]

    def __init__(self):
        self.app = self.create_app()
        self.code = None
        self.srv = None

        self.disable_logging()
        self.start_server()

    @classmethod
    def disable_logging(cls):
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

    def create_app(self):
        from flask import Flask
        app = Flask(__name__)

        @app.route('/admin/auth/init')
        def auth_init():
            self.code = request.args.get('code')
            return redirect("/admin/auth/finish")

        @app.route('/admin/auth/finish')
        def auth_finish():
            self.srv.shutdown_signal = True
            return "You can go now back to the console"

        return app

    def run(self):
        self.srv.serve_forever()

    @property
    def redirect_uri(self):
        return 'http://localhost:%s/admin/auth/init' % self.srv.port

    def start_server(self):
        ports = self.ports[:]
        random.shuffle(ports)

        for port in ports:
            try:
                self.srv = make_server('127.0.0.1', port, self.app)
                break
            except OSError:
                continue


def wait_for_input():
    if sys.version_info >= (3, 0):
        code = input('Enter the token ')
    else:
        # noinspection PyUnresolvedReferences
        code = raw_input('Enter the token ')

    return code


# noinspection PyUnusedLocal
def pixy_flow(ctx):
    verifier = base64url(os.urandom(32))
    verifier_challenge = base64url(sha256(verifier))

    verifier = verifier.decode('ascii')
    verifier_challenge = verifier_challenge.decode('ascii')

    app = LocalWebServer() if ctx.local_web_server else None

    redirect_uri = app.redirect_uri if app is not None else ctx.host + '/admin/auth/init'
    query = {
        'response_type': 'code',
        'scope': 'openid offline_access user_external_id',
        'client_id': ctx.client_id,
        'redirect_uri': redirect_uri,
        'code_challenge': verifier_challenge,
        'code_challenge_method': 'S256'
    }

    authorize_url = 'https://{}/authorize?{}'.format(auth0_url(ctx.auth0), url_encode(query))

    print("If the browser doesn't open enter this URL manually\n%s\n" % authorize_url)

    webbrowser.open(authorize_url)

    if app is not None:
        app.run()
        code = app.code
    else:
        code = wait_for_input()

    r = requests.post('https://{}/oauth/token'.format(auth0_url(ctx.auth0)), json={
        'code': code,
        'code_verifier': verifier,
        'client_id': ctx.client_id,
        'grant_type': 'authorization_code',
        'redirect_uri': redirect_uri,
    })

    r.raise_for_status()

    data = r.json()

    print("Success!, you sre authorized to use the command line.")

    return data['access_token'], data['refresh_token'], data['id_token']


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options
